/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.3 $";
#endif

/*
 * AIX SCSI functions
 */

#include "defs.h"
#if	defined(HAVE_SYS_SCSI_H)
#include <sys/scsi.h>
#endif
#if	defined(HAVE_SYS_SCDISK_H)
#include <sys/scdisk.h>
#endif
#if	defined(HAVE_SYS_CDROM_H)
#include <sys/cdrom.h>
#endif
#if	defined(HAVE_SYS_TAPE_H)
#include <sys/tape.h>
#endif
#if	defined(HAVE_SYS_TMSCSI_H)
#include <sys/tmscsi.h>
#endif
#include "myscsi.h"

/*
 * XXX Disabled SCSI query on AIX since it rarely works since openx() only
 * works if the device is not in use and because of reports of system
 * SCSI errors on AIX 4.3 systems on the rare occasion when it does work.
 */
#undef HAVE_STRUCT_SC_IOCMD

/*
 * Issue a SCSI command and return the results.
 */
extern int ScsiCmd(ScsiCmd, Query)
     ScsiCmd_t		       *ScsiCmd;
     ScsiQuery_t	       *Query;
{
#if	defined(HAVE_STRUCT_SC_IOCMD)
    int				IoctlCmd = 0;
    char                       *IoctlCmdStr = NULL;
    static char			Buff[SCSI_BUF_LEN];
    static struct sc_iocmd	Cmd;
    ScsiCdbG0_t		       *CdbPtr = NULL;
    int				Status;
    size_t			DataLen = 0;
    char		       *What = NULL;

    if (!Query || !Query->DevInfo || 
	!ScsiCmd || !ScsiCmd->Cdb || !ScsiCmd->CdbLen || 
	ScsiCmd->DevFD < 0 || !ScsiCmd->DevFile) {
	SImsg(SIM_DBG, "ScsiCmd: Bad parameters.");
	return(-1);
    }

    /* Get the length of the data buffer we're going to get back */
    CdbPtr = (ScsiCdbG0_t *) ScsiCmd->Cdb;
    DataLen = CdbPtr->Length;
    if (DataLen > sizeof(Buff)) {
	SImsg(SIM_DBG, "ScsiCmd(): Sizeof Cdb Length (%d) > Buff size (%d)",
	      DataLen, sizeof(Buff));
	return(-1);
    }

    switch (Query->DevInfo->Type) {
#if	defined(DKIOCMD)
    case DT_DISKDRIVE:
	IoctlCmd = DKIOCMD;
	IoctlCmdStr = "DKIOCMD";
	break;
#endif
#if	defined(CDIOCMD)
    case DT_CDROM:
	IoctlCmd = CDIOCMD;
	IoctlCmdStr = "CDIOCMD";
	break;
#endif
#if	defined(STIOCMD)
    case DT_TAPEDRIVE:
	IoctlCmd = STIOCMD;
	IoctlCmdStr = "STIOCMD";
	break;
#endif
    default:
#if	defined(TMIOCMD)
	IoctlCmd = TMIOCMD;
	IoctlCmdStr = "TMIOCMD";
	break;
#endif
    }

    if (!IoctlCmd) {
	SImsg(SIM_DBG, "%s: No SCSI ioctl command defined for device type %s",
	      ScsiCmd->DevFile, TypeGetType(Query->DevInfo->Type));
	return -1;
    } else {
	SImsg(SIM_DBG, "%s: Using ioctl command %s",
	      ScsiCmd->DevFile, IoctlCmdStr);
    }

    (void) memset(Buff, 0, sizeof(Buff));
    (void) memset(&Cmd, 0, sizeof(Cmd));

    (void) memcpy((void *)Cmd.scsi_cdb, (void *)ScsiCmd->Cdb,
		  sizeof(Cmd.scsi_cdb));
    Cmd.command_length = ScsiCmd->CdbLen;
    Cmd.buffer = (caddr_t) Buff;
    Cmd.data_length = DataLen;
    Cmd.timeout_value = MySCSI_CMD_TIMEOUT;	/* timeout_value==seconds */
    Cmd.flags = B_READ; /* | SC_ASYNC; */

    /*
     * Send cmd to device
     */
    errno = 0;
    Status = ioctl(ScsiCmd->DevFD, IoctlCmd, &Cmd);
    if (Status < 0 || Cmd.status_validity != 0) {
	/* 
	 * We just need Cdb.cmd so it's ok to assume ScsiCdbG0_t here
	 */
	CdbPtr = (ScsiCdbG0_t *) ScsiCmd->Cdb;
	switch (errno) {
	case EACCES:
	    What = "device not in Diagnostic mode (EACCES)";
	    break;
	case EINVAL:
	    What = "data buffer size too large (EINVAL)";
	    break;
	default:
	    What = SYSERR;
	}
	SImsg(SIM_GERR, "%s: SCSI command 0x%x failed: %s", 
	      ScsiCmd->DevFile, CdbPtr->Cmd, What);
	if (Cmd.status_validity)
	    SImsg(SIM_GERR, 
		  "%s: SCSI command 0x%x failed: validity=0x%X bus_status=0x%X adapter_status=0x%X",
		  ScsiCmd->DevFile, CdbPtr->Cmd, Cmd.status_validity,
		  Cmd.scsi_bus_status, Cmd.adapter_status);
	return(-1);
    }

    ScsiCmd->Data = (void *) Buff;
    return(0);
#else	/* ! HAVE_STRUCT_SC_IOCMD */
    return(-1);
#endif	/* HAVE_STRUCT_SC_IOCMD */
}
