/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.94 $";
#endif

/*
 * SGI IRIX specific functions
 *
 * Basic device information is derived via getinvent(3) and reverse engineering
 * the IRIX hinv(1m) command which uses getinvent(3).  Many parts of the
 * device code found here remains untested.  Some code is #ifdef disabled
 * because I don't know how some #define's in some files (like sys/invent.h)
 * coorespond to "code" names in other .h files.  This is especially true
 * of the graphics code.
 * 
 * Special Graphics information is derived from getinvent(3), and the
 * "gfx_info" structure, various graphics device specific *.h files
 * and reverse engineering the IRIX /usr/gfx/gfxinfo command.
 */

#include "defs.h"
#include "myirixdeclare.h"
#include "myscsi.h"
#include <mntent.h>
#include <sys/systeminfo.h>
#include <sys/conf.h>
#include <sys/stat.h>
#include <sys/syssgi.h>
#include <sys/dsreq.h>
#include "myirixdeclare.h"

/*
 * Disk
 */
#include <sys/dkio.h>
#include <sys/dvh.h>

/*
 * Platform Specific Interfaces
 */
extern Opaque_t				GetCpuInfoIRIX();
extern Opaque_t			        GetModelInv();
extern Opaque_t			        GetSerialSGI();
extern Opaque_t			        GetNumCpuSGI();
extern Opaque_t				GetKernVerSGI();
extern Opaque_t			        GetOSVerSGI();
extern Opaque_t 		        GetManLongSysinfoSGI();
extern Opaque_t			        GetVirtMemSGI();
extern Opaque_t			        BuildSoftInfoInst();
extern Opaque_t			        BuildDevices();

PsiSwitch_t PsiSwitch = {
    /* GetBootTime */	{ GetBootTimeUtmp, NULL },
    /* GetCpuInfo */	{ GetCpuInfoIRIX, NULL },
    /* GetCpuType */ 	{ GetCpuTypeCpuInfo, GetCpuTypeSGI, NULL },
    /* GetAppArch */ 	{ GetAppArchSysinfo, NULL },
    /* GetKernArch */ 	{ GetKernArchSysinfo, NULL },
    /* GetKernBits */ 	{ GetKernBitsOsName, NULL },
    /* GetKernVer */ 	{ GetKernVerSGI, NULL },
    /* GetLibcName */ 	{ GetLibcNameShared, NULL },
    /* GetLibcVer */ 	{ GetLibcVersionShared, NULL },
    /* GetManLong */ 	{ GetManLongSysinfoSGI, GetManLongDef, NULL },
    /* GetManShort */ 	{ GetManShortSysinfo, GetManShortDef, NULL },
    /* GetMemory */ 	{ GetMemoryPhysmemSym, NULL },
    /* GetVirtMem */ 	{ GetVirtMemSwapctl, GetVirtMemSGI, NULL },
    /* GetModel */ 	{ GetModelInv, GetModelFile, GetModelDef, NULL },
    /* GetNumCpu */ 	{ GetNumCpuSGI, NULL },
    /* GetOsDist */ 	{ NULL },
    /* GetOsName */ 	{ GetOSNameUname, NULL },
    /* GetOsVer */ 	{ GetOSVerSGI, GetOSVerSysinfo, NULL },
    /* GetRomVer */  	{ NULL },
    /* GetSerial */ 	{ GetSerialSGI, NULL },
    /* BuildDevices */  { BuildDevices, NULL },
    /* BuildSoftInfo */ { BuildSoftInfoInst, NULL },
    /* BuildPartInfo */ { BuildPartInfoDevices, NULL },
    /* BuildPrinter */	{ BuildPrinterLp, NULL },
};

/*
 * Get CpuInfo_t IRIX style
 */
extern Opaque_t GetCpuInfoIRIX()
{
    static CpuInfo_t	       *CpuInfo;
    DevInfo_t		       *DevTree = NULL;
    DevInfo_t		       *DevInfo;
    static DevFind_t		Find;

    SImsg(SIM_DBG, "GetCpuInfoIRIX: Called");

    if (CpuInfo)
	return CpuInfo;

    /* First build a full device tree */
    BuildDevices(&DevTree, NULL);

    /* Look for the first CPU dev we can find */
    SImsg(SIM_DBG, "GetCpuInfoIRIX: Searching for a DT_CPU");
    DevFindInit(&Find);
    Find.Tree = DevTree;
    Find.DevType = DT_CPU;
    if (DevInfo = DevFind(&Find))
	CpuInfo = (CpuInfo_t *) DevInfo->DevSpec;
    else
	SImsg(SIM_DBG, "GetCpuInfoIRIX: Cannot find any DT_CPU devices.");

    return CpuInfo;
}

/*
 * Get serial number using the syssgi() system call.
 */
extern Opaque_t GetSerialSGI()
{
    static char			Serial[MAXSYSIDSIZE];
    register char	       *cp;
    int				Okay;

    if (Serial[0])
	return(Serial);

    if (syssgi(SGI_SYSID, Serial) == 0) {
	/*
	 * Silly kludge to see if Serial contains valid information.
	 */
	Okay = TRUE;
	for (cp = Serial; cp && *cp; ++cp)
	    if (!isalnum(*cp)) {
		Okay = FALSE;
		break;
	    }
	if (Okay)
	    return(Serial);
    } else
	SImsg(SIM_GERR, "syssgi(SGI_SYSID, ...) failed: %s", SYSERR);

    return(GetSerialSysinfo());
}

/*
 * SGI Get Kernel Version
 */
extern Opaque_t GetKernVerSGI()
{
    static char 		Buff[128];
    char 			Version[256];
    long 			Style = -1;
  
#if	defined(_SC_KERN_POINTERS)
    Style = sysconf(_SC_KERN_POINTERS);
#endif
    sysinfo(SI_VERSION, Version, sizeof(Version));
    /* IRIX Release 6.0 IP21 Version 08241804 System V - 64 Bit */
    snprintf(Buff, sizeof(Buff),
	     "IRIX Release %s %s Version %s System V (%d bit)",
	     GetOSVerSGI(), GetKernArchSysinfo(),
	     Version, (Style == -1) ? 32 : Style);

    return( (char *) Buff);
}

/*
 * SGI Get OS Version number
 */
extern Opaque_t GetOSVerSGI()
{
    static char 		Build[256];
    char 			Patch[50];	/* To hold patch version */
    char 		       *pos = Build;
    char 			MinVer[50];
    int				Len;

    if (sysinfo(_MIPS_SI_OSREL_MAJ, Build, sizeof(Build)) < 0) {
	(void) snprintf(Build, sizeof(Build), "%s", GetOSVerUname());
	return (char *) Build;
    } else
	pos += strlen(Build);

    if (sysinfo(_MIPS_SI_OSREL_MIN, MinVer, sizeof(MinVer)) >= 0) {
	if (MinVer[0] != '.')
	    *pos++ = '.';
	Len = pos - &Build[0];
	(void) snprintf(pos, sizeof(Build)-Len, "%s", MinVer);

	/* Pre-load buffer with an initial '.', then stuff in version after */
	if (sysinfo(_MIPS_SI_OSREL_PATCH, Patch+1, sizeof(Patch)) >= 0) {
	    Patch[0] = '.';	/* Pre-load string with a '.' delimiter */
	    if ( ! ((strlen(Patch) == 2) && (Patch[1] == '0')) ) {
		Len = pos - Build;
		(void) snprintf(pos, sizeof(Build)-Len, "%s", Patch);
	    }
	}
    }

    if (pos == Build)		/* If we didn't get anything, return NULL */
	return (char *) NULL;
    else
	return (char *) Build;
}

/*
 * Get the CPU manufacturer name.
 */
extern char *GetCpuMfgSGI()
{
    static char		        Buff[64];

    if (Buff[0])
	return Buff;

    if (sysinfo(SI_ARCHITECTURE, Buff, sizeof(Buff)) >= 0)
	return Buff;

    return (char *) NULL;
}

/*
 * Get the CPU model name.
 */
extern char *GetCpuModelSGI()
{
    static char		        Buff[64];
    char		       *cp;

    if (Buff[0])
	return Buff;

    if (sysinfo(_MIPS_SI_PROCESSORS, Buff, sizeof(Buff)) >= 0) {
	if (cp = strchr(Buff, ','))
	    *cp = CNULL;
	return Buff;
    }

    return (char *) NULL;
}

/*
 * Determine CPU type using sysinfo().
 */
extern Opaque_t GetCpuTypeSGI()
{
    char		       *Type = NULL;
    char		       *cp;

    if (!Type)
	Type = GetCpuModelSGI();

    if (!Type)
	Type = GetCpuMfgSGI();

    if (Type)
	if (cp = strchr(Type, ' '))
	    *cp = CNULL;

    return Type;
}

/*
 * Get Number of CPU's using sysinfo()
 */
extern Opaque_t GetNumCpuSGI()
{
    static char 		Buff[32];

    if (sysinfo(_MIPS_SI_NUM_PROCESSORS, Buff, sizeof(Buff)) < 0)
	return (char *) NULL;
    else
	return Buff;
}

/*
 * Get the long manufacterers name
 */
extern Opaque_t GetManLongSysinfoSGI()
{
    static char 		Buff[128];

    if (sysinfo(_MIPS_SI_VENDOR, Buff, sizeof(Buff)) < 0)
	return (char *) NULL;
    else
	return (char *) Buff;
}

/*
 * Get size of virtual memory.
 */
extern Opaque_t GetVirtMemSGI()
{
    kvm_t		       *kd;
    long			AvailMem;
    long			UsedMem;
    off_t			Amount = 0;
    nlist_t		       *nlPtr;

    if (kd = KVMopen()) {
	if ((nlPtr = KVMnlist(kd, "availsmem", (nlist_t *)NULL, 0)) == NULL)
	    return((char *) NULL);

	if (CheckNlist(nlPtr))
	    return((char *) NULL);

	if (KVMget(kd, nlPtr->n_value, (char *) &AvailMem,
		   sizeof(AvailMem), KDT_DATA) >= 0)
	    Amount = AvailMem;

	if ((nlPtr = KVMnlist(kd, "usermem", (nlist_t *)NULL, 0)) == NULL)
	    return((char *) NULL);

	if (CheckNlist(nlPtr))
	    return((char *) NULL);

	if (KVMget(kd, nlPtr->n_value, (char *) &UsedMem,
		   sizeof(UsedMem), KDT_DATA) >= 0)
	    Amount += UsedMem;

	Amount = Amount * (GETPAGESIZE() / KBYTES);
	KVMclose(kd);
    }

    return(GetVirtMemStr(Amount));
}

/*
 * Initialize the OS specific parts of the Device Types table
 */
void DevTypesInit()
{
}

/*
 * Issue a SCSI command and return the results.
 * Uses the ds(7M) interface.
 */
extern int ScsiCmd(ScsiCmd)
     ScsiCmd_t		       *ScsiCmd;
{
#if	defined(DS_ENTER)
    static char			Buff[SCSI_BUF_LEN];
    static dsreq_t		Cmd;
    ScsiCdbG0_t		       *CdbPtr = NULL;

    if (!ScsiCmd || !ScsiCmd->Cdb || !ScsiCmd->CdbLen || 
	ScsiCmd->DevFD < 0 || !ScsiCmd->DevFile) {
	SImsg(SIM_DBG, "ScsiCmd: Bad parameters.");
	return(-1);
    }

    (void) memset(Buff, 0, sizeof(Buff));
    (void) memset(&Cmd, 0, sizeof(Cmd));

    Cmd.ds_cmdbuf = (caddr_t) ScsiCmd->Cdb;
    Cmd.ds_cmdlen = ScsiCmd->CdbLen;
    Cmd.ds_databuf = (caddr_t) Buff;
    Cmd.ds_datalen = sizeof(Buff);
    Cmd.ds_time = MySCSI_CMD_TIMEOUT * 1000;	/* ds_time==milliseconds */
    Cmd.ds_flags = DSRQ_READ;

    /* 
     * We just need Cdb.cmd so it's ok to assume ScsiCdbG0_t here
     */
    CdbPtr = (ScsiCdbG0_t *) ScsiCmd->Cdb;

    /*
     * Send cmd to device
     */
    if (ioctl(ScsiCmd->DevFD,  DS_ENTER, &Cmd) == -1) {
	SImsg(SIM_GERR, "%s: ioctl DS_ENTER for SCSI 0x%x failed: %s", 
	      ScsiCmd->DevFile, CdbPtr->Cmd, SYSERR);
	return(-1);
    }

    if (Cmd.ds_status != 0) {
	SImsg(SIM_GERR, "%s: SCSI command 0x%x failed with ds_status=%d", 
	      ScsiCmd->DevFile, CdbPtr->Cmd, Cmd.ds_status);
	return(-1);
    }

    ScsiCmd->Data = (void *) Buff;
#endif	/* DS_ENTER */
    return(0);
}

/*
 * Clean a string of unprintable characters and excess white-space.
 */
static char *CleanStr(String, StrSize)
    char		       *String;
    int				StrSize;
{
    register int		i, n;
    char		       *NewString;

    NewString = (char *) xcalloc(1, StrSize + 1);

    for (i = 0, n = 0; i < StrSize; ++i) {
	if (i == 0)
	    /* Skip initial white space */
	    while (isspace(String[i]))
		++i;
	/* Skip extra white space */
	if (i > 0 && isspace(String[i-1]) && isspace(String[i]))
	    continue;
	if (isprint(String[i]))
	    NewString[n++] = String[i];
    }

    /* Remove trailing white space */
    while (isspace(NewString[n]))
	NewString[n--] = CNULL;

    return(NewString);
}

/*
 * Check to see if MntDevice is one of the "special" ones
 * that IRIX uses.
 */
static int IsSpecialUse(MntDevice, DevPath)
    char		       *MntDevice;
    char		       *DevPath;
{
    static struct stat		MntStat;
    static struct stat		DevStat;

    if (!EQ(MntDevice, "/dev/root") &&
	!EQ(MntDevice, "/dev/usr") &&
	!EQ(MntDevice, "/dev/swap"))
	return(0);

    if (stat(MntDevice, &MntStat) != 0) {
	SImsg(SIM_GERR, "stat failed: %s: %s", MntDevice, SYSERR);
	return(0);
    }
    if (stat(DevPath, &DevStat) != 0) {
	SImsg(SIM_GERR, "stat failed: %s: %s", DevPath, SYSERR);
	return(0);
    }

    if (MntStat.st_rdev == DevStat.st_rdev)
	return(1);
    else
	return(0);
}

/*
 * Look for a partition named DevName in the mount table.
 */
static int GetMountInfo(PartInfo, FilePtr)
    PartInfo_t		       *PartInfo;
    FILE		       *FilePtr;
{
    struct mntent              *MntEnt;
    char		      **Argv;

    while (MntEnt = getmntent(FilePtr)) {
	if (EQ(MntEnt->mnt_fsname, PartInfo->DevPath) || 
	    IsSpecialUse(MntEnt->mnt_fsname, PartInfo->DevPath)) {
	    if (EQ(MntEnt->mnt_type, MNTTYPE_SWAP))
		PartInfo->Usage = PIU_SWAP;
	    else
		PartInfo->Usage = PIU_FILESYS;
	    if (MntEnt->mnt_type)
		PartInfo->Type = strdup(MntEnt->mnt_type);
	    if (MntEnt->mnt_dir)
		PartInfo->MntName = strdup(MntEnt->mnt_dir);
	    if (MntEnt->mnt_opts && !EQ(MntEnt->mnt_opts, "-"))
		if (StrToArgv(MntEnt->mnt_opts, ",", &Argv, NULL, 0) > 0)
		    PartInfo->MntOpts = Argv;
	    return 0;
	}
    }

    return -1;
}

/*
 * Lookup the usage for a given partition.
 */
static int PartUseLookup(PartInfo)
     PartInfo_t		       *PartInfo;
{
    static FILE 	       *mntFilePtr = NULL;
    static FILE 	       *fstabFilePtr = NULL;

    if (!PartInfo)
	return -1;

    /*
     * First try the current mount table
     */
    if (!mntFilePtr) {
	if ((mntFilePtr = setmntent(MOUNTED, "r")) == NULL) {
	    SImsg(SIM_GERR, "%s: Cannot open for reading: %s.", 
		  MNTTAB, SYSERR);
	    return -1;
	}
    } else
	rewind(mntFilePtr);

    if (GetMountInfo(PartInfo, mntFilePtr) == 0)
	return 0;

    /*
     * Now try the static mount table (/etc/fstab),
     * which may not reflect current reality.
     */
    if (!fstabFilePtr) {
	if ((fstabFilePtr = setmntent(MNTTAB, "r")) == NULL) {
	    SImsg(SIM_GERR, "%s: Cannot open for reading: %s.", 
		  MNTTAB, SYSERR);
	    return -1;
	}
    } else
	rewind(fstabFilePtr);

    if (GetMountInfo(PartInfo, fstabFilePtr) == 0)
	return 0;

    return -1;
}

/*
 * Get amount/space used for this partition
 */
#include <sys/statvfs.h>
static int GetPartAmtUsed(Part)
     PartInfo_t		       *Part;
{
    static struct statvfs	Statvfs;
    Large_t			u;

    if (!Part || !Part->MntName)
	return -1;

    if (statvfs(Part->MntName, &Statvfs) != 0) {
	SImsg(SIM_DBG, "%s: statvfs failed: %s", Part->MntName, SYSERR);
	return -1;
    }

    u = MBYTES / (Large_t)Statvfs.f_frsize;
    Part->AmtUsed = (Large_t) (Statvfs.f_blocks - Statvfs.f_bfree) / u;

    return 0;
}

/*
 * Find out what a partition is being used for.
 */
static void GetPartUse(PartInfo)
     PartInfo_t		       *PartInfo;
{
    char			Type[128];
    char			Usage[128];

    PartUseLookup(PartInfo);
    if (PartInfo->Usage && PartInfo->Type)
	return;

    Type[0] = Usage[0] = CNULL;

    switch (PartInfo->TypeNum) {
#if	defined(PTYPE_VOLHDR)
    case PTYPE_VOLHDR:	
	strcpy(Type, "volhdr");		strcpy(Usage, "Volume Header");	
	break;
#endif
#if	defined(PTYPE_TRKREPL)
    case PTYPE_TRKREPL:	
	strcpy(Type, "trkrepl");	strcpy(Usage, "Track Replacement");
	break;
#endif
#if	defined(PTYPE_SECREPL)
    case PTYPE_SECREPL:	
	strcpy(Type, "secrepl");	strcpy(Usage, "Sector Replacement");
	break;
#endif
#if	defined(PTYPE_LVOL)
    case PTYPE_LVOL:	
	strcpy(Type, "lvol");		strcpy(Usage, "Logical Volume");
	break;
#endif
#if	defined(PTYPE_RLVOL)
    case PTYPE_RLVOL: 
	strcpy(Type, "rlvol");		strcpy(Usage, "Raw Logical Volume");
	break;
#endif
#if	defined(PTYPE_VOLUME)
    case PTYPE_VOLUME:	
	strcpy(Type, "volume");		strcpy(Usage, "Entire Volume");
	break;
#endif
#if	defined(PTYPE_XLV)
    case PTYPE_XLV:
	strcpy(Type, "xlv");		strcpy(Usage, "XLV Volume");
	break;
#endif
#if	defined(PTYPE_XFSLOG)
    case PTYPE_XFSLOG:
	strcpy(Type, "xfslog");		strcpy(Usage, "XFS Log");
	break;
#endif

	/*
	 * These types of partitions should be looked up
	 */
#if	defined(PTYPE_RAW)
    case PTYPE_RAW:	strcpy(Type, "raw");			break;
#endif
#if	defined(PTYPE_EFS)
    case PTYPE_EFS:	strcpy(Type, "efs");			break;
#endif
#if	defined(PTYPE_SYSV)
    case PTYPE_SYSV:	strcpy(Type, "sysv");			break;
#endif
#if	defined(PTYPE_BSD)
    case PTYPE_BSD:	strcpy(Type, "bsd");			break;
#endif
#if	defined(PTYPE_XFS)
    case PTYPE_XFS:	strcpy(Type, "xfs");			break;
#endif

    default:		strcpy(Type, "UNKNOWN");		break;
    }

    if (Type[0] && !PartInfo->Type)
	PartInfo->Type = strdup(Type);

    if (Usage[0] && !PartInfo->Usage)
	PartInfo->UsageStatus = strdup(Usage); 
}

/*
 * Get Disk Partition information for a given disk
 */
static DiskPart_t *GetDiskPart(DevInfo, VolHdr)
    DevInfo_t		       *DevInfo;
    struct volume_header       *VolHdr;
{
    DiskPart_t		       *New = NULL;
    DiskPart_t		       *Last = NULL;
    DiskPart_t		       *Base = NULL;
    PartInfo_t		       *PartInfo = NULL;
    static char			Path[MAXPATHLEN];
    register int		i;

    for (i = 0; i < NPARTAB; ++i) {
	if (VolHdr->vh_pt[i].pt_nblks == 0)
	    continue;

	PartInfo = PartInfoCreate(NULL);

	(void) snprintf(Path, sizeof(Path),  "s%d", i);
	PartInfo->Name = strdup(Path);

	(void) snprintf(Path, sizeof(Path),  "%s/%s%s", _PATH_DEV_DSK,
			DevInfo->Name, PartInfo->Name);
	PartInfo->DevPath = strdup(Path);

	(void) snprintf(Path, sizeof(Path),  "%s/%s%s", _PATH_DEV_RDSK,
			DevInfo->Name, PartInfo->Name);
	PartInfo->DevPathRaw = strdup(Path);

	PartInfo->Num = i;

	PartInfo->StartSect = (Large_t) VolHdr->vh_pt[i].pt_firstlbn;
	PartInfo->NumSect = (Large_t) VolHdr->vh_pt[i].pt_nblks;
	/* Store pt_type for possible later use in finding drive capacity */
	PartInfo->TypeNum = VolHdr->vh_pt[i].pt_type;
	/* What's it used for? */
	GetPartUse(PartInfo);
	/* How much space is used? */
	GetPartAmtUsed(PartInfo);

	New = NewDiskPart(NULL);
	New->PartInfo = PartInfo;
	if (Last) {
	    Last->Next = New;
	    Last = New;
	} else
	    Base = Last = New;
    }

    return(Base);
}

/*
 * Get total disk size (capacity) in bytes.
 * Look for the PTYPE_VOLUME (entire disk) partiton and use its 
 * size as return value.
 */
static float IRIXGetDiskSize(DiskDrive)
    DiskDrive_t		       *DiskDrive;
{
    DiskPart_t		       *dp;

#if	defined(PTYPE_VOLUME)
    for (dp = DiskDrive->DiskPart; dp; dp = dp->Next) {
	if (dp->PartInfo && dp->PartInfo->TypeNum == PTYPE_VOLUME) {
	    SImsg(SIM_DBG, "DISK PTYPE_VOLUME numsects=%.0f size=%.0f",
		  (float) dp->PartInfo->NumSect, 
		  (float) DiskDrive->SecSize);
	    return (float) ((float)dp->PartInfo->NumSect * 
			    (float)DiskDrive->SecSize);
	}
    }
#else
    SImsg(SIM_DBG, "PTYPE_VOLUME is not defined on this OS");
#endif	/* PTYPE_VOLUME */

    return((float) 0);
}

/*
 * Query disk using SCSI methods
 */
extern int ProbeDiskDriveScsi(DevInfo, DevFile)
     DevInfo_t		       *DevInfo;
     char		       *DevFile;
{
    int				fd;

    fd = open(DevFile, O_RDONLY|O_NDELAY|O_NONBLOCK);
    if (fd < 0) {
	SImsg(SIM_GERR, "open failed: %s: %s", DevFile, SYSERR);
	return(-1);
    }

    (void) ScsiQuery(DevInfo, DevFile, fd, TRUE);

    (void) close(fd);

    return(0);
}

/*
 * Query disk using normal OS methods
 */
extern int ProbeDiskDriveQuery(DevInfo, DevFile)
     DevInfo_t		       *DevInfo;
     char		       *DevFile;
{
    DiskDriveData_t	       *DiskDriveData = NULL;
    DiskDrive_t		       *OSdisk = NULL;
    DiskDrive_t		       *HWdisk = NULL;
    char		       *Vendor = NULL;
    u_int			DiskCap = 0;
    static struct volume_header	VolHdr;
    static char			DriveType[SCSI_DEVICE_NAME_SIZE + 1];
    int				fd;

    fd = open(DevFile, O_RDONLY);
    if (fd < 0) {
	SImsg(SIM_GERR, "open failed: %s: %s", DevFile, SYSERR);
	return(-1);
    }

    if (!DevInfo->Model || !DevInfo->Vendor) {
	/*
	 * Get the Drive Type
	 */
	if (ioctl(fd, DIOCDRIVETYPE, DriveType) < 0)
	    SImsg(SIM_GERR, "ioctl DIOCDRIVETYPE failed: %s: %s", 
		  DevFile,SYSERR);
	else {
	    if (Vendor = strchr(DriveType, ' ')) {
		*Vendor = CNULL;
		++Vendor;
	    } else
		Vendor = DriveType;
	    if (!DevInfo->Vendor)
		DevInfo->Vendor = strdup(Vendor);
	    if (!DevInfo->Model)
		DevInfo->Model = strdup(DriveType);
	}
    }

    DiskDriveData = (DiskDriveData_t *) DevInfo->DevSpec;
    OSdisk = DiskDriveData->OSdata;
    HWdisk = DiskDriveData->HWdata;

    /*
     * Get the volume header data.
     */
    if (ioctl(fd, DIOCGETVH, &VolHdr) < 0) {
	SImsg(SIM_GERR, "ioctl DIOCGETVH failed: %s: %s", DevFile, SYSERR);
    } else {
	OSdisk->DiskPart = GetDiskPart(DevInfo, &VolHdr);
	OSdisk->SecSize = VolHdr.vh_dp.dp_secbytes;
#if	OSVER <= 63
	/*
	 * This stuff was removed starting with IRIX 6.4
	 */
	OSdisk->DataCyl = VolHdr.vh_dp.dp_cyls;
	OSdisk->Tracks = VolHdr.vh_dp.dp_trks0;
	OSdisk->Sect = VolHdr.vh_dp.dp_secs;
	OSdisk->IntrLv = VolHdr.vh_dp.dp_interleave;
	OSdisk->APC = VolHdr.vh_dp.dp_spares_cyl;
#endif	/* OSVER <= 63 */
	if (isprint(VolHdr.vh_bootfile[0]))
	    AddDevDesc(DevInfo, VolHdr.vh_bootfile, "Boot File", 0);
    }

#if	defined(DIOCREADCAPACITY)
    /*
     * Get the Disk Capacity
     */
    if (ioctl(fd, DIOCREADCAPACITY, &DiskCap) < 0)
	SImsg(SIM_GERR, "ioctl DIOCREADCAPACITY failed: %s: %s", 
	      DevFile, SYSERR);
    else {
	SImsg(SIM_DBG, "DISK: %s DIOCREADCAPACITY = %d", DevFile, DiskCap);
	/*
	 * If we still don't have the disk capacity, try getting it
	 * from the partition table.
	 */
	DiskCap = (u_int) IRIXGetDiskSize(OSdisk);
	if (HWdisk && !HWdisk->Size && DiskCap && OSdisk && OSdisk->SecSize) {
	    HWdisk->Size = (float) nsect_to_mbytes(DiskCap, OSdisk->SecSize);
	    SImsg(SIM_DBG, "DISK: %s: size=%.0f MB", 
		  DevInfo->Name, OSdisk->Size);
	}
    }
#endif	/* DIOCREADCAPACITY */

    (void) close(fd);

    return(0);
}

/*
 * Build device tree using TreePtr.
 * Calls bus and method specific functions to
 * search for devices.
 */
extern Opaque_t BuildDevices(TreePtr, Names)
    DevInfo_t 		       **TreePtr;
    char		       **Names;
{
    int 			 Found = 1;
    static DevInfo_t		*SavedTreePtr;

    if (SavedTreePtr) {
	*TreePtr = SavedTreePtr;
	return (Opaque_t) 0;
    }

    if (BuildDevicesInvent(TreePtr, Names) == 0)
	Found = 0;
#if	defined(BUILD_HWG)
    if (BuildDevicesHwg(TreePtr, Names) == 0)
	Found = 0;
#endif	/* BUILD_HWG */

    if (TreePtr)
	SavedTreePtr = *TreePtr;

    return (Opaque_t) Found;
}
