/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.6 $";
#endif

/*
 * Linux SCSI specific functions
 */

#include "defs.h"
#include "strutil.h"
#include "linux-declare.h"
#include "myscsi.h"
#if	defined(HAVE_SCSI_SCSI_H)
#	include <scsi/scsi.h>
#else	/* ! HAVE_SCSI_SCSI_H */
# 	if 	defined(HAVE__USR_SRC_LINUX_INCLUDE_SCSI_SCSI_H)
#		include "/usr/src/linux/include/scsi/scsi.h"
#	endif	/* HAVE__USR_SRC_LINUX_INCLUDE_SCSI_SCSI_H */
#endif	/* HAVE_SCSI_SCSI_H */
#if	defined(HAVE_SCSI_SG_H)
#	include <scsi/sg.h>
#	define HAVE_SCSI_SUPPORT
#else	/* !HAVE_SCSI_SCSI_H */
# 	if 	defined(HAVE__USR_SRC_LINUX_INCLUDE_SCSI_SG_H)
#		include "/usr/src/linux/include/scsi/sg.h"
#		define HAVE_SCSI_SUPPORT
#	endif	/* HAVE__USR_SRC_LINUX_INCLUDE_SCSI_SG_H */
#endif	/* HAVE_SCSI_SCSI_H */

/*
 * Issue a SCSI command and return the results.  This function uses
 * the SCSI Generic (sg) interface.  See 
 * 	http://www.linuxdoc.org/HOWTO/SCSI-Programming-HOWTO.html
 */
extern int ScsiCmd(ScsiCmd)
     ScsiCmd_t		       *ScsiCmd;
{
#if	defined(HAVE_SCSI_SUPPORT)
#define SG_OFFSET		sizeof(struct sg_header)
    static char			Buff[SG_OFFSET + SCSI_BUF_LEN];
    int				Status;
    int				Timeout;
    ScsiCdbG0_t		       *CdbPtr = NULL;
    struct sg_header		ScsiHeader;

    if (!ScsiCmd || !ScsiCmd->Cdb || !ScsiCmd->CdbLen || 
	ScsiCmd->DevFD < 0 || !ScsiCmd->DevFile) {
	SImsg(SIM_DBG, "ScsiCmd: Bad parameters.");
	return -1;
    }

    /*
     * Set the timeout
     */
    Timeout = MySCSI_CMD_TIMEOUT * 1000;	/* timeout == milliseconds */
    if (ioctl(ScsiCmd->DevFD, SG_SET_TIMEOUT, &Timeout) < 0) {
	SImsg(SIM_DBG, "%s: ScsiCmd: Set SG Timeout to %d msecs failed: %s",
	      ScsiCmd->DevFile, Timeout, SYSERR);
    }

    (void) memset(Buff, 0, sizeof(Buff));

    /*
     * SCSI Header for sg
     */
    (void) memset(&ScsiHeader, 0, sizeof(ScsiHeader));
    ScsiHeader.reply_len = sizeof(Buff);
    ScsiHeader.twelve_byte = ScsiCmd->CdbLen == 12;
    (void) memcpy(Buff, &ScsiHeader, sizeof(ScsiHeader));

    /*
     * Add Command Data Block
     */
    (void) memcpy(Buff + SG_OFFSET, ScsiCmd->Cdb, ScsiCmd->CdbLen);

    /* 
     * We just need Cdb.cmd so it's ok to assume ScsiCdbG0_t here
     */
    CdbPtr = (ScsiCdbG0_t *) ScsiCmd->Cdb;

    /*
     * Issue SCSI Command
     */
    Status = write(ScsiCmd->DevFD, Buff, SG_OFFSET + ScsiCmd->CdbLen);
    if (Status < 0 || Status != SG_OFFSET + ScsiCmd->CdbLen || 
	ScsiHeader.result) {
	SImsg(SIM_GERR, 
	      "%s: SCSI command issue 0x%x failed: %s",
	      ScsiCmd->DevFile, CdbPtr->Cmd, SYSERR);
	return -1;
    }

    /*
     * Read results of command
     */
    Status = read(ScsiCmd->DevFD, Buff, sizeof(Buff));
    if (Status < 0) {
	SImsg(SIM_GERR, 
	      "%s: SCSI command read command 0x%x failed: %s",
	      ScsiCmd->DevFile, CdbPtr->Cmd, SYSERR);
	return -1;
    }
	
    ScsiCmd->Data = (void *) Buff + SG_OFFSET;

#endif	/* HAVE_SCSI_SUPPORT */

    return 0;
}

/*
 * Query a SCSI device with base (driver) named Name and Unit.
 */
static DevList_t *ScsiProbe(Name, Unit, DevType, DevInfo)
     char		       *Name;
     int			Unit;
     int			DevType;
     DevInfo_t		       *DevInfo;
{
    DevInfo_t		     *(*ProbeFunc)() = NULL;
    DevList_t		       *DevList;
    static ProbeData_t		ProbeData;
    static char			DevName[64];
    static char			File[128];
    static int			CDUnits;	/* # CDs found */
    static int			DiskUnits;	/* # Disks found */
    static int			TapeUnits;	/* # Tapes found */
    int				fd;
    struct Scsi_Id {
	long	l1; /* target | lun << 8 | channel << 16 | low_ino << 24 */
	long	l2; /* Unique id */
    } 				Scsi_Id;

    (void) snprintf(DevName, sizeof(DevName), "%s%c", Name, Unit + 'a');
    (void) snprintf(File, sizeof(File), "%s%s", _PATH_DEV, DevName);

    /*
     * Only use O_RDWR as other flags affect sg driver
     */
    fd = open(File, O_RDWR);
    if (fd < 0) {
	SImsg(SIM_DBG, "%s: open %s failed: %s", DevName, File, SYSERR);
	if (errno != EBUSY)
	    return((DevList_t *) NULL);
    }

    SImsg(SIM_DBG, "%s:\tSCSI Probe: Device Exists", File);

    /*
     * Create the DevInfo
     */
    if (!DevInfo)
	DevInfo = NewDevInfo(NULL);
    if (!DevInfo->Name)
	DevInfo->Name = strdup(DevName);
    if (!DevInfo->Driver)
	DevInfo->Driver = strdup(Name);
    if (!DevInfo->ClassType)
	DevInfo->ClassType = CT_SCSI;
    if (!DevInfo->Type)
	DevInfo->Type = DevType;
    DevAddFile(DevInfo, strdup(File));

    /*
     * Crate DevList and return it.
     */
    DevList = (DevList_t *) xcalloc(1, sizeof(DevList_t));
    DevList->Name = DevInfo->Name;
    DevList->File = DevInfo->Files[0];
    DevList->DevInfo = DevInfo;

    /*
     * Identify SCSI specific ID of this device now
     */
    if (ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &DevList->Bus) < 0) {
	SImsg(SIM_DBG, "%s: ioctl to GET_BUS_NUMER failed: %s",
	      File, SYSERR);
    }
    if (ioctl(fd, SCSI_IOCTL_GET_IDLUN, &Scsi_Id) < 0) {
	SImsg(SIM_DBG, "%s: ioctl to GET_IDLUN failed: %s",
	      File, SYSERR);
    }
    DevInfo->Unit = DevList->Unit = Scsi_Id.l1 & 0xFF;
    DevList->Lun = (Scsi_Id.l1 >> 8L) & 0xFF;
    SImsg(SIM_DBG, "%s:\tSCSI Bus=%d Unit=%d LUN=%d",
	  File, DevList->Bus, DevList->Unit, DevList->Lun);

    /*
     * Do full SCSI query of this device
     */
    (void) ScsiQuery(DevInfo, File, fd, TRUE);

    /*
     * Now that Type is known, set the canonical device name properly.
     */
    DevName[0] = CNULL;
    switch (DevInfo->Type) {
    case DT_DISKDRIVE:
	(void) snprintf(DevName, sizeof(DevName), "sd%c", (DiskUnits++) + 'a');
	ProbeFunc = ProbeDiskDrive;
	break;
    case DT_TAPEDRIVE:
	(void) snprintf(DevName, sizeof(DevName), "st%d", TapeUnits++);
	break;
    case DT_CD:
	(void) snprintf(DevName, sizeof(DevName), "scd%d", CDUnits++);
	break;
    }
    if (DevName[0]) {
	DevInfo->AltName = DevInfo->Name;
	DevInfo->Name = strdup(DevName);
	(void) snprintf(File, sizeof(File), "%s%s", _PATH_DEV, DevName);
	DevAddFile(DevInfo, strdup(File));
    }
    if (ProbeFunc) {
        (void) memset(&ProbeData, CNULL, sizeof(ProbeData));
	ProbeData.DevName = DevInfo->Name;
	ProbeData.DevFile = File;
	ProbeData.FileDesc = fd;
	ProbeData.UseDevInfo = DevInfo;
        (void) (*ProbeFunc)(&ProbeData);
    }

    (void) close(fd);

    return(DevList);
}

/*
 * Detect SCSI devices by checking every /dev/sg? device.
 * This method causes unnecessary probes which can slow us down.
 * However, this is much more desirable than using /proc/scsi/scsi
 * since this method insures we get the proper mapping of OS Device
 * Name -> /dev/sgX.  This is especially true on systems with more than
 * 1 SCSI controller.
 */
extern DevList_t *ScsiDetect()
{
    register int		u;
    static DevList_t	       *ScsiDevices;
    static DevList_t	       *Last;
    register DevList_t	       *New = NULL;

    if (ScsiDevices)
	return(ScsiDevices);

    /*
     * Look for SCSI devices via generic SG driver
     */
    for (u = 0; u < MAX_SCSI_UNITS; ++u) {
	if (New = ScsiProbe("sg", u, 0, NULL)) {
	    if (!ScsiDevices)
		ScsiDevices = Last = New;
	    else {
		Last->Next = New;
		Last = New;
	    }
	}
    }

    return(ScsiDevices);
}

#if notdef
/*
 * XXX We don't use this version of ScsiDetect() because it's nearly
 * impossible to map the SCSI Unit # -> OS Device Name (i.e. "sdb") ->
 * SCSI Generic driver device /dev/sgX (i.e. "sgb"), especially on
 * systems with more than 1 SCSI controller.
 */
/*
 * Detect SCSI devices by reading /proc/scsi/scsi.  Most Linux kernels
 * (including 2.2 - 2.4) support this file.
 *
 * This method is good in that we only have to check the system for the devices
 * which the OS has already discovered.  It's bad in that this interface
 * was never intended to be easily parsed by software.
 */
extern DevList_t *ScsiDetectProc()
{
    register int		u;
    static DevList_t	       *ScsiDevices;
    static DevList_t	       *Last;
    register DevList_t	       *New = NULL;
    DevInfo_t		       *DevInfo;
    DevType_t		       *dt;
    FILE		       *fp;
    static char			Line1[512];
    static char			Line2[512];
    static char			Line3[512];
    char		      **Argv;
    int				Argc;
    char		       *Type;
    char		       *cp;

    if (ScsiDevices)
	return(ScsiDevices);

    SImsg(SIM_DBG, "ScsiDetect: Reading <%s>", PROC_FILE_SCSI);

    fp = fopen(PROC_FILE_SCSI, "r");
    if (!fp) {
	SImsg(SIM_DBG, "%s: Open for reading failed: %s",
	      PROC_FILE_SCSI, SYSERR);
	return (DevList_t *) NULL;
    }

    /* 
     * Read and disgard first line which should be:
     * 	Attached devices:
     */
    (void) fgets(Line1, sizeof(Line1), fp);
    if (cp = strchr(Line1, '\n'))
	*cp = CNULL;
    SImsg(SIM_DBG, "ScsiDetect: First line <%s>", Line1);

    /*
     * There are 3 sequential lines per device so we read 1 device
     * set at a go.
     */
    while (fgets(Line1, sizeof(Line1), fp) &&
	   fgets(Line2, sizeof(Line2), fp) &&
	   fgets(Line3, sizeof(Line3), fp)) {
	New = NULL;
	DevInfo = NewDevInfo(NULL);

	if (cp = strchr(Line1, '\n'))
	    *cp = CNULL;
	if (cp = strchr(Line2, '\n'))
	    *cp = CNULL;
	if (cp = strchr(Line3, '\n'))
	    *cp = CNULL;
	SImsg(SIM_DBG, "ScsiDetect: Line1=<%s>", Line1);
	SImsg(SIM_DBG, "ScsiDetect: Line2=<%s>", Line2);
	SImsg(SIM_DBG, "ScsiDetect: Line3=<%s>", Line3);

	/*
	 * Parse first line which looks like:
	 *	Host: scsi0 Channel: 00 Id: 04 Lun: 00
	 */
	Argc = StrToArgv(Line1, " ", &Argv, NULL, STRTO_SKIPRP);
	if (Argc < 8) {
	    SImsg(SIM_DBG, "ScsiDetect: Line1=<%s> Arg count wrong (%d)",
		  Line1, Argc);
	    continue;
	}
	if (Argv[5])
	    DevInfo->Unit = atoi(Argv[5]);	/* SCSI HW Unit # */

	/*
	 * Parse the second line which looks like:
	 *  Vendor: IBM      Model: DNES-318350W     Rev: SA30
	 */
	/* Skip leading whitespace */
	for (cp = Line2; isspace(*cp); ++cp);
	Argc = StrToArgv(cp, " ", &Argv, NULL, STRTO_SKIPRP);
	if (Argc < 6) {
	    SImsg(SIM_DBG, "ScsiDetect: Line2=<%s> Arg count wrong (%d)",
		  cp, Argc);
	    continue;
	}
	if (Argv[1])
	    DevInfo->Vendor = strdup(Argv[1]);
	if (Argv[3])
	    DevInfo->Model = strdup(Argv[3]);
	if (Argv[5])
	    DevInfo->Revision = strdup(Argv[5]);

	/*
	 * Parse the third line which looks like:
	 *  Type:   Direct-Access                    ANSI SCSI revision: 03
	 */
	/* Skip leading whitespace */
	for (cp = Line3; isspace(*cp); ++cp);
	Argc = StrToArgv(cp, " ", &Argv, NULL, STRTO_SKIPRP);
	if (Argc < 6) {
	    SImsg(SIM_DBG, "ScsiDetect: Line3=<%s> Arg count wrong (%d)",
		  cp, Argc);
	    continue;
	}
	Type = Argv[1];

	/*
	 * Determine the device type
	 */	
	if (EQ(Type, "Direct-Access"))
	    DevInfo->Type = DT_DISKDRIVE;
	else if (EQ(Type, "Sequential-Access"))
	    DevInfo->Type = DT_TAPEDRIVE;
	else if (EQ(Type, "CD-ROM"))
	    DevInfo->Type = DT_CDROM;
	else if (EQ(Type, "Printer"))
	    DevInfo->Type = DT_PRINTER;
	else if (EQ(Type, "Scanner"))
	    DevInfo->Type = DT_SCANNER;
	else if (dt = TypeGetByName(Type))	/* Look it up in our table */
	    DevInfo->Type = dt->Type;
	else
	    SImsg(SIM_DBG, "%s: Unit %d (%s/%s) is an unknown type = <%s>",
		  PROC_FILE_SCSI, DevInfo->Unit,
		  PRTS(DevInfo->Vendor), PRTS(DevInfo->Model), PRTS(Type));

	SImsg(SIM_DBG, 
	      "ScsiDetect: FOUND Unit=%d Vendor=<%s> Model=<%s> Type=<%s>",
	      DevInfo->Unit, PRTS(DevInfo->Vendor), PRTS(DevInfo->Model),
	      Type);

	/* Probe device now */
	if (DevInfo->Type)
	    New = ScsiProbe("sg", DevInfo->Unit, DevInfo->Type, DevInfo);

	/* Add device to list */
	if (New) {
	    if (!ScsiDevices)
		ScsiDevices = Last = New;
	    else {
		Last->Next = New;
		Last = New;
	    }
	}
    }

    return(ScsiDevices);
}
#endif	/* notdef */
