/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.6 $";
#endif

#include "defs.h"
#include "dmi.h"

static DevDesc_t	      **lp;

/*
 * Values for Processor Family (DmiStructProc_t.Type)
 * See section 3.3.5.2 of SMBIOS Ref Spec
 */
static NameTab_t ProcFamilies[] = {
    { 0x03, "8086" },
    { 0x04, "80286" },
    { 0x05, "Intel386" },
    { 0x06, "Intel486" },
    { 0x07, "8087" },
    { 0x08, "80287" },
    { 0x09, "80387" },
    { 0x0A, "80487" },
    { 0x0B, "Pentium" },
    { 0x0C, "Pentium Pro" },
    { 0x0D, "Pentium II" },
    { 0x0E, "Pentium MMX" },
    { 0x0F, "Celeron" },
    { 0x10, "Pentium II Xeon" },
    { 0x11, "Pentium III" },
    { 0x12, "M1" },
    { 0x19, "K5" },
    { 0x20, "Power PC" },
    { 0x21, "Power PC 601" },
    { 0x22, "Power PC 603" },
    { 0x23, "Power PC 603+" },
    { 0x24, "Power PC 604" },
    { 0x30, "Alpha" },		/* Sometimes means Pentium Pro */
    { 0x40, "MIPS" },
    { 0x50, "SPARC" },
    { 0x60, "68040" },
    { 0x61, "68xxx" },
    { 0x62, "680000" },
    { 0x63, "68010" },
    { 0x64, "68020" },
    { 0x65, "68030" },
    { 0x70, "Hobbit" },
    { 0x80, "Weitek" },
    { 0x90, "PA-RISC" },
    { 0xA0, "V30" },
    { 0xB0, "Pentium III Xeon" },
    { 0, NULL },
};

/*
 * Decode DMI Processor Information (DMI_STRUCT_PROC)
 */
extern DevInfo_t *DmiDecodeProc(Hdr, Data, Dmi, Query)
     DmiHeader_t	       *Hdr;
     void		       *Data;
     Dmi_t		       *Dmi;
     DmiQuery_t		       *Query;
{
    DevInfo_t		       *DevInfo;
    CpuInfo_t		       *CpuInfo;
    DmiStructProc_t	       *Proc;
    char		       *Mfg = NULL;
    char		       *Model = NULL;
    char		       *Version = NULL;
    char		       *Serial = NULL;
    char		       *Asset = NULL;
    char		       *Socket = NULL;
    char		       *Part = NULL;
    char		       *cp;
    char		       *Type;
    int				Class;
    static char			Buff[128];
    static int			Unit;

    Proc = (DmiStructProc_t *) Data;

    if ((cp = DmiString(Hdr, Proc->Mfg)) && !DmiIsGeneric(cp))
	Mfg = cp;
    if (cp = GetNameTab(Proc->Family, ProcFamilies))
	Model = cp;
    if ((cp = DmiString(Hdr, Proc->Version)) && !DmiIsGeneric(cp))
	Version = cp;
    if ((cp = DmiString(Hdr, Proc->Socket)) && !DmiIsGeneric(cp))
	Socket = cp;
    if (Dmi->SmbiosVersion >= 2030 && Hdr->Length > 0x20) {
	if ((cp = DmiString(Hdr, Proc->Serial)) && !DmiIsGeneric(cp))
	    Serial = cp;
	if ((cp = DmiString(Hdr, Proc->AssetTag)) && !DmiIsGeneric(cp))
	    Asset = cp;
	if ((cp = DmiString(Hdr, Proc->Part)) && !DmiIsGeneric(cp))
	    Part = cp;
    }

    SImsg(SIM_DBG, 
	  "DMI PROC: Mfg=<%s> Family=<%s> Version=<%s>",
	  PRTS(Mfg), PRTS(Model), PRTS(Version));
    SImsg(SIM_DBG, 
	  "DMI PROC:\tsizeof(DmiStructProc_t)=%d (0x%X) Hdr->Length=%d (0x%X)",
	  sizeof(DmiStructProc_t), sizeof(DmiStructProc_t), 
	  Hdr->Length, Hdr->Length);
    SImsg(SIM_DBG, 
	  "DMI PROC:\tSerial=<%s> Asset=<%s> Part=<%s> Socket=<%s>",
	  PRTS(Serial), PRTS(Asset), PRTS(Part), PRTS(Socket));
    SImsg(SIM_DBG, 
	  "DMI PROC:\tClock=%d Speed=%d/%d Upgrade=0x%02X",
	  Proc->ExtClockFreq, Proc->CurSpeed, Proc->MaxSpeed, Proc->Upgrade);
    SImsg(SIM_DBG, 
	  "DMI PROC:\tCache Handles: L1=0x%04X L2=0x%04X L3=0x%04X",
	  Proc->L1handle, Proc->L2handle, Proc->L3handle);

    /*
     * A 0 for current speed indicates the CPU isn't present.
     * DMI is telling us this probably because it's an unused CPU slot.
     */
    if (Proc->CurSpeed <= 0) {
	SImsg(SIM_DBG,
	      "DMI PROC:\tCurrent Speed=%d means the CPU isn't present.",
	      Proc->CurSpeed);
	return (DevInfo_t *) NULL;
    }

    /*
     * If this is a specific query, then set the values now and return
     */
    if (Query && Query->Type == Hdr->Type) {
	Query->ValStr1 = Mfg;
	Query->ValStr2 = Model;
	Query->ValStr3 = Version;
	Query->ValStr4 = Serial;
	Query->ValStr5 = Asset;
	Query->ValInt1 = (Proc->CurSpeed) ? Proc->CurSpeed : Proc->MaxSpeed;
	return DevInfo;
    }

    DevInfo = NewDevInfo(NULL);
    DevInfo->Type = DT_PROCESSOR;
    DevInfo->Unit = Unit++;
    Type = NULL;
    Class = 0;
    switch (Proc->Type) {
    case DMI_PROC_TYPE_CPU:	Type = "cpu";	Class = CT_CPU; 	break;
    case DMI_PROC_TYPE_MATH:	Type = "math";	Class = CT_MATH;	break;
    case DMI_PROC_TYPE_DSP:	Type = "dsp";	Class = CT_DSP;		break;
    case DMI_PROC_TYPE_VIDEO:	Type = "videoproc"; Class = CT_VIDEO;	break;
    default:
	Type = "processor";
	SImsg(SIM_DBG, "DmiDecodeProc: Unknown Processor Type=%d",
	      Proc->Type);
    }
    snprintf(Buff, sizeof(Buff), "%s%d", Type, DevInfo->Unit);
    DevInfo->Name = strdup(Buff);
    DevInfo->ClassType = Class;

    lp = &(DevInfo->DescList);

    if (Mfg)		DevInfo->Vendor = Mfg;
    if (Model)		DevInfo->Model = Model;
    if (Version)	DevInfo->Revision = Version;
    if (Serial)		DevInfo->Serial = Serial;
    if (Part)		DevInfo->Part = Part;
    if (Proc->ExtClockFreq)
	AddDesc(lp, "FS Bus Speed", 
		GetClockStr((Large_t)Proc->ExtClockFreq, (Large_t)MHERTZ), 0);
    if (Asset)
	AddDesc(lp, "Asset Tag", Asset, 0);
    if (Socket)
	AddDesc(lp, "Socket", Socket, 0);

    if (Class == CT_CPU)
	if (CpuInfo = (CpuInfo_t *) GetCpuInfoX86()) {
	    if (Proc->CurSpeed)
		CpuInfo->Speed = (Large_t) Proc->CurSpeed;
	    else if (Proc->MaxSpeed)
		CpuInfo->Speed = (Large_t) Proc->MaxSpeed;
	    DevInfo->DevSpec = (Opaque_t) CpuInfo;
	}

    if (Dmi->SmbiosVersion >= 2010) {
	if (Proc->L1handle)
	    DmiGetCache(Proc->L1handle, 1, Dmi, Query, DevInfo);
	if (Proc->L2handle)
	    DmiGetCache(Proc->L2handle, 2, Dmi, Query, DevInfo);
	if (Proc->L3handle)
	    DmiGetCache(Proc->L3handle, 3, Dmi, Query, DevInfo);
    }

    return DevInfo;
}
