#
# $Id: uiLicense.pm,v 1.5 2002/03/26 23:50:10 mcooper Exp $
#
# SysInfo UI License functions
#

package uiLicense;

require 5.003;
require Exporter;

use mcDebug;
use uiMisc;
use uiGlobals;
use uiWindow;
use uiKVF;

my $LicenseTermsAns;
my $LicTermsWin;

BEGIN {
    #
    # Export function
    #
    @ISA      = qw(Exporter);

    #
    # Variables and functions being auto exported
    #
    @EXPORT   = qw(
		   LicenseInfo LicenseTermsView LicenseCheck
		   LicenseTermsAsk LicenseTermsShow
		  );
}

#
# Bind a Message to something usually a botton
#
sub BindStatus {
    my ($Widget, $Msg) = @_;

    $Widget->bind('<Enter>', [ sub { $StatusMessage = $_[1] }, $Msg ]);
    $Widget->bind('<Leave>', sub { $StatusMessage = ""; });
}

#
# Display license info
#
sub LicenseInfo {
    if (!$SysInfoPath) {
	Error("The \$SysInfoPath variable is not set.");
	return -1;
    }

    my $FH = GetFH();
    my $Cmd = sprintf("%s -license", $SysInfoPath);
    if (!open($FH, "$Cmd 2>&1 |")) {
	Error("Cannot run <%s> command: %s", $Cmd, $!);
	return -1;
    }

    my @Contents = <$FH>;
    close($FH);
    if ($#Contents < 0) {
	Error("No input found for <%s>.", $Cmd);
	return -1;
    }
    WindowDisplay({Contents=>\@Contents, Title=>$Title . ' License Info',
		   Image=>$SplashLogo});

    return 0;
}

#
# Check license status and display warning if necessary
#
sub LicenseCheck {
    my $Args = shift;
    my $SysInfoPath = $Args->{SysInfoPath};

    my $FH = GetFH();
    my $Cmd = sprintf("%s -license", $SysInfoPath);
    if (!open($FH, "$Cmd 2>&1 |")) {
	Error("Cannot run <%s> command: %s", $Cmd, $!);
	return -1;
    }

    my @Contents = <$FH>;
    close($FH);
    if ($#Contents < 0) {
	Error("No input found for <%s>.", $Cmd);
	return -1;
    }
    my $IsDemo = 0;
    foreach my $Line (@Contents) {
	if (lc($Line) =~ /license type:/ && lc($Line) =~ /demo/) {
	    $IsDemo = 1;
	    last;
	}
    }
    if (!$IsDemo) {
	return 0;
    }

    WindowDisplay({Contents=>\@Contents, Title=>$Title . ' License Check',
		   Image=>$SplashLogo});

    return 0;
}

#
# Get license terms Prefs key
#
sub LicenseTermsPrefsKey {
    my $Version = BEversion();

    my $Key = 'AcceptLicense_' . $Version;

    return $Key;
}

#
# Process License Terms Accept
#
sub LicenseTermsAccept {
    Debug("LicenseTermsAccept()");

    $LicTermsWin->withdraw();
    # Save results
    my $Key = LicenseTermsPrefsKey();
    $Prefs->{$Key} = 1;
    KVFsave({File=>$PrefsFile, KV=>$Prefs});

    # XXX This is unreliable
    # $LicTermsWin->grabRelease();

    return 0;
}

#
# Process License Terms Decline
#
sub LicenseTermsDecline {
    Debug("LicenseTermsDecline()");
    Quit();
}

#
# Show license terms
#
sub LicenseTermsShow {
    my $Args = shift;
    my $Prefs = $Args->{Prefs};

    if ($Args->{Ask}) {
	my $Key = LicenseTermsPrefsKey();
	if ($Prefs->{$Key}) {
	    Debug("AcceptLicenseCheck: Already accepted.");
	    return 0;
	}
    }

    my @Contents = ReadFile({File=>$LicenseTermsFile});
    if ($#Contents < 0) {
	Error("No license terms file content found.");
	return -1;
    }

    $LicTermsWin = MainWindow->new(-title=>WinTitle('License Terms'));
    my $Fr = $LicTermsWin->Frame(-borderwidth=>5,
				 -relief=>'groove')
	->pack(-side=>'top', -anchor=>'center');

    my $Msg = $Fr->Scrolled("Text",
			    -scrollbars=>'osoe',
			    -font=>$FontText,
			    -background=>'white',
			    -height=>30,
			    -width=>80)
	->pack(-side=>'top', -anchor=>'center');

    $Msg->tagConfigure('center', -justify=>'center');
    foreach my $Line (@Contents) {
	$Msg->insert('end', $Line);
    }

    if ($Args->{Ask}) {
	my $ab = $Fr->Button(-text=>'Accept',
			     -font=>$FontButton,
			     -command=>\&LicenseTermsAccept)
	    ->pack(-side=>'left', -expand=>1, -anchor=>'center',
		   -padx=>10, -pady=>10);
	BindStatus($ab, 'Accept license terms');
	my $db = $Fr->Button(-text=>'Decline',
			     -font=>$FontButton,
			     -command=>\&LicenseTermsDecline)
	    ->pack(-side=>'left', -expand=>1, -anchor=>'center',
		   -padx=>10, -pady=>10);
	BindStatus($db, 'Decline license terms');

	# Grab keyboard+mouse from other windows
	# XXX This is unreliable
	##$LicTermsWin->grab();
    } else {
	$Fr->Button(-text=>'Close',
		    -font=>$FontButton,
		    -command=>sub { $LicTermsWin->withdraw })
	    ->pack(-side=>'left', -expand=>1, -anchor=>'center',
		   -padx=>10, -pady=>10);
    }

    return 0;
}

#
# View License Terms
#
sub LicenseTermsView {
    my $Args = shift;

    $Args->{View} = 1;
    LicenseTermsShow($Args);
}

#
# Check to see if user has accepted license yet
#
sub LicenseTermsAsk {
    my $Args = shift;

    $Args->{Ask} = 1;
    LicenseTermsShow($Args);
}

1;
